#! /usr/bin/env python
""" Distribution file for x/84. """
from __future__ import print_function
import subprocess
import pipes
import errno
import sys
import os

# pylint: disable=E0611,F0401
#         No name 'core' in module 'distutils'
#         Unable to import 'distutils.core'
from distutils.core import setup
from setuptools.command.develop import develop as _develop
from setuptools import Command

HERE = os.path.dirname(__file__)
README = 'README.rst'
DOC_URL = 'http://x84.rtfd.org'


def check_virtualenv():
    """ Ensure a virtualenv is used. """
    if not os.getenv('VIRTUAL_ENV'):
        print('You must be in a virtualenv, See developer documentation '
              'at filepath docs/developers.rst or online at {0}'
              .format(DOC_URL), file=sys.stderr)
        exit(1)


class Develop(_develop):

    """ Ensure a virtualenv is used and install developer requirements. """

    def finalize_options(self):
        """ Validate options. """
        check_virtualenv()
        _develop.finalize_options(self)

    def run(self):
        """ Run develop command. """
        cargs = ['pip', 'install', '--upgrade', 'sphinx', 'tox']
        print('>>', ' '.join(map(pipes.quote, list(cargs))))
        subprocess.check_call(cargs)
        _develop.run(self)


class BuildDocs(Command):

    """ Build documentation using sphinx. """

    #: path to source documentation folder.
    DOCS_SRC = os.path.join(HERE, 'docs')

    #: path to output html documentation folder.
    DOCS_DST = os.path.join(HERE, 'build', 'docs')

    user_options = []

    def initialize_options(self):
        """ Initialize options. """
        pass

    def finalize_options(self):
        """ Validation options. """
        # pylint: disable=R0201
        #         Method could be a function
        check_virtualenv()

    def run(self):
        """ Call sphinx-build. """
        try:
            subprocess.call(
                ('sphinx-build', '-v', '-n', '-E',
                 self.DOCS_SRC, self.DOCS_DST),
                stdout=sys.stdout, stderr=sys.stderr)
        except OSError as err:
            if err.errno != errno.ENOENT:
                raise
            print("run '{0} develop' first".format(__file__), file=sys.stderr)
            sys.exit(1)


setup(name='x84',
      version='2.0.10',
      description=("Framework for Telnet and SSH BBS or MUD server "
                   "development with example default bbs board"),
      long_description=open(os.path.join(HERE, README)).read(),
      author='Jeff Quast',
      author_email='contact@jeffquast.com',
      url=DOC_URL,
      keywords="telnet ssh terminal server ansi bbs mud curses utf8 cp437",
      license='ISC',
      packages=['x84', 'x84.default', 'x84.default.art', 'x84.bbs',
                'x84.encodings', 'x84.webmodules'],
      package_data={
          '': [README],
          'x84.default': ['*.ans', '*.txt', ],
          'x84.default.art': ['*.asc', '*.ans', '*.txt',
                              'top/*',
                              'weather/*',
                              'bulletins/*/*',
                              ],
      },
      install_requires=[
          'blessed==1.9.5',
          'requests==2.5.1',
          'irc==11.0.1',
          'sqlitedict==1.1.0',
          'python-dateutil==2.3',
          'jaraco.timing==1.1',
          'jaraco.util==10.6',
          'more-itertools==2.2',
          'sauce==1.1',
          'six==1.8.0',
          'wsgiref==0.1.2',
          'xmodem==0.3.2',
          'feedparser==5.1.3',
          'html2text==2014.12.29',
      ],
      extras_require={
          'with_crypto': (
              # These cryptogaphy requirements may only be installed:
              # - if a C compiler is available,
              # - if libssl is available,
              # - (sometimes, only) if libffi is available
              #
              # for this reason, they are **optional**, so that x/84 may be
              # installed without a compiler or these external C libraries
              # -- however it is **highly** recommended to always try to
              # use install x84[with_crypto].  It has always been a goal for
              # x/84 to be "pure python" to remain compatible with alternative
              # python interpreter implementations.
              'bcrypt==1.1.0',
              'cffi==0.8.6',
              'cryptography==0.7.1',
              'ecdsa==0.11',
              'enum34==1.0.4',
              'paramiko==1.15.2',
              'pyOpenSSL==0.14',
              'pyasn1==0.1.7',
              'pycparser==2.10',
              'pycrypto==2.6.1',
              'web.py==0.37',
              'cherrypy==3.6.0',
          )
      },
      entry_points={
          'console_scripts': ['x84=x84.engine:main'],
      },
      classifiers=[
          'Environment :: Console :: Curses',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: ISC License (ISCL)',
          'Natural Language :: English',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: POSIX :: BSD :: FreeBSD',
          'Operating System :: POSIX :: BSD :: NetBSD',
          'Operating System :: POSIX :: BSD :: OpenBSD',
          'Operating System :: POSIX :: BSD',
          'Operating System :: POSIX :: Linux',
          'Operating System :: POSIX :: SunOS/Solaris',
          'Operating System :: POSIX',
          'Operating System :: Unix',
          'Programming Language :: Python :: 2 :: Only',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Topic :: Artistic Software',
          'Topic :: Communications :: BBS',
          'Topic :: Software Development :: User Interfaces',
          'Topic :: Terminals :: Telnet',
          'Topic :: Terminals',
      ],
      cmdclass={
          'develop': Develop,
          'docs': BuildDocs,
      },
      zip_safe=False,
)



""" File browser/manager for x/84. """
# std imports
from __future__ import division
import zipfile
import os

# local
from x84.bbs import getsession, getterminal, echo, getch, syncterm_setfont
from x84.bbs import get_ini, DBProxy, Lightbar, LineEditor
from x84.bbs import send_modem, recv_modem
from x84.default.common import filesize

#: file description database
DIZ_DB = 'filediz'

#: root folder for the filebase
ROOT = get_ini(section='sftp', key='root') or '/usr/share/misc'

#: directory for storing uploads
UPLOADS_DIR = os.path.join(ROOT, '__uploads__')

#: color used for menu key entries
COLOR_HIGHLIGHT = get_ini(
    section='fbrowse', key='color_highlight'
) or 'bold_white_on_blue'

#: color used for borders
COLOR_BORDER = get_ini(
    section='fbrowse', key='color_border',
) or 'bold_blue'

#: fontset for SyncTerm emulator
SYNCTERM_FONT = get_ini(
    section='fbrowse', key='syncterm_font'
) or 'cp437'

#: marker for flagged files in browser list
FLAGGED_CHAR = get_ini(
    section='fbrowse', key='flagged_char'
) or u'+'

#: extensions for ASCII collies
COLLY_EXTENSIONS = get_ini(
    section='fbrowse', key='colly_extensions', split=True
) or ['.txt', '.asc', '.ans']

#: name of virtual directory for collecting user's flagged files
FLAGGED_DIRNAME = get_ini(
    section='fbrowse', key='flagged_dirname'
) or u'__flagged__{0}'.format(os.path.sep)

#: decoding to use for ASCII collies
COLLY_DECODING = get_ini(
    section='fbrowse', key='colly_decoding'
) or 'amiga'


class FileBrowser(object):

    """ File browsing interface variables that need to be passed around """

    # pylint:disable=R0903
    diz_location = 0
    last_diz_len = 0
    max_diz_width = 0
    max_diz_height = 0
    flagged_files = set()
    diz_extractors = dict()

# instance to be used throughout the script
browser = FileBrowser()  # pylint:disable=C0103


def diz_from_dms(binary, filename):
    """
    Amiga diskmasher format. Depends on the external binary 'xdms'.
    """
    import subprocess
    args = ('d', filename)
    proc = subprocess.Popen((binary,) + args, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    output, _ = proc.communicate()
    if proc.returncode == 0:
        return output.decode('cp437_art')
    else:
        return u'No description'


def diz_from_lha(binary, filename):
    """
    Amiga LHA format. Depends on the external binary 'lha'.
    """
    import subprocess
    import tempfile
    import shutil
    description = u'No description'
    path = tempfile.mkdtemp(prefix='x84_')
    args = ('xw={0}'.format(path), filename)
    proc = subprocess.Popen((binary,) + args, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    proc.wait()
    dizfilename = os.path.join(path, 'file_id.diz')
    if proc.returncode == 0 and os.path.isfile(dizfilename):
        with open(dizfilename, 'rb') as dizfile:
            description = dizfile.read().decode('cp437_art')
    try:
        shutil.rmtree(path)
    except OSError:
        pass
    return description


def diz_from_zip(filename, method=zipfile.ZIP_STORED):
    """
    Pull FILE_ID.DIZ from `filename` using particular zipfile `method`.
    """
    try:
        myzip = zipfile.ZipFile(filename, compression=method, allowZip64=True)
        for cname in (cname for cname in myzip.namelist()
                      if cname.lower() == 'file_id.diz'):
            return myzip.read(cname).decode('cp437_art')
        return u'No description'
    except zipfile.BadZipfile:
        return u'Bad zip file, cannot parse'
    except zipfile.LargeZipFile:
        # since we do allowZip64=True above, this shouldn't happen any longer
        return u'Large zip file, cannot parse'
    except NotImplementedError:
        return u'Unsupported compression, cannot parse'


def get_diz_from_colly(filepath):
    """ Get FILE_ID.DIZ from within an ASCII collection. """
    colly = open(filepath, 'r').read()
    colly_diz_begin = '@BEGIN_FILE_ID.DIZ'
    colly_diz_end = '@END_FILE_ID.DIZ'
    pos = colly.find(colly_diz_begin)
    if pos > 0:
        colly = colly[pos + len(colly_diz_begin):]
        pos = colly.find(colly_diz_end)
        if pos > 0:
            return colly[:pos].splitlines()
    return None


def get_instructions(term, is_sysop=None):
    """ Show file browser instructions. """
    return [
        term.bold_blue_underline(u'Instructions'),
        u' ',
        u'{0} Un/flag file for download'
        .format(term.reverse(u'(SPACE)')),
        u'{0}  Back up'
        .format(term.reverse(u'(LEFT)')),
        term.reverse(u'(RIGHT, ENTER)'),
        u'        Browse subdirectory',
        u'{0}     Download flagged file(s)'
        .format(term.reverse(u'(D)')),
        u'{0}     Unflag all files'
        .format(term.reverse(u'(-)')),
        u'{0}     Upload file(s)'
        .format(term.reverse(u'(u)')),
        u'{0}     Edit description'
        .format(term.reverse(u'(e)')) if is_sysop else u'',
        u'{0}     Quit'
        .format(term.reverse(u'(q)')),
        u' ',
        u'Files are also available via {0}'
        .format(term.bold(u'SFTP')),
    ]


def edit_description(filepath, db_desc):
    """ Edit file description. """
    from x84.bbs import gosub
    new_desc = None
    if filepath in db_desc:
        new_desc = u'\r\n'.join([line.decode('cp437_art')
                                 for line in db_desc[filepath]])
    new_desc = gosub('editor', continue_draft=new_desc)
    if not new_desc:
        return
    with db_desc:
        db_desc[filepath] = new_desc.splitlines()


def download_files(term, session, protocol='xmodem1k'):
    """ Download flagged files. """
    if not len(browser.flagged_files):
        return False
    echo(term.clear)
    flagged = browser.flagged_files.copy()
    for fname in flagged:
        _fname = fname[fname.rfind(os.path.sep) + 1:].decode('utf8')
        echo(term.bold_green(
            u'Start your {protocol} receiving program '
            u'to begin transferring {_fname}...\r\n'
            .format(protocol=protocol, _fname=_fname)))
        echo(u'Press ^X twice to cancel\r\n')

        fin = open(fname, 'rb')
        if not send_modem(fin, protocol):
            echo(term.bold_red(u'Transfer failed!\r\n'))
        else:
            browser.flagged_files.remove(fname)
            session.user['flaggedfiles'] = browser.flagged_files
            echo(term.bold(u'Transfer(s) finished.\r\n'))
    term.inkey()


def upload_files(term, protocol='xmodem1k'):
    """ Upload files. """
    echo(term.clear)
    while True:
        echo(u'Filename (empty to quit):\r\n')
        led = LineEditor(width=term.width - 1)
        led.refresh()
        inp = led.read()
        led = None
        if inp:
            for illegal in (os.path.sep, u'..', u'~',):
                if illegal in inp:
                    echo(term.bold_red(u'\r\nIllegal filename.\r\n'))
                    term.inkey()
                    return

            echo(term.bold(
                u'\r\nBegin your {0} sending program now.\r\n'
                .format(protocol)))

            upload_filename = os.path.join(UPLOADS_DIR, inp)
            try:
                upload = open(upload_filename, 'wb')
            except IOError as err:
                echo(term.bold_red('u\r\nIOError: {err}\r\n'.format(err=err)))
            else:
                if not recv_modem(upload, protocol):
                    echo(term.bold_red(u'Upload failed!\r\n'))
                    os.unlink(upload_filename)
                else:
                    echo(term.bold_green(u'Transfer succeeded.\r\n'))
            term.inkey()
        else:
            return


def draw_interface(term, lightbar):
    """ Redraw and resize the interface. """
    lightbar.height = term.height
    lightbar.width = max(10, int(term.width * 0.25))
    # +1 for spacing between lightbar and diz
    browser.diz_location = lightbar.width + 1
    # -2 for spacing between the lightbar, the diz, and the edge of the screen
    browser.max_diz_width = term.width - lightbar.width - 2
    # -4 for space above/below diz area and info line (filename, size)
    browser.max_diz_height = term.height - 4
    echo(u''.join([term.clear,
                   lightbar.border(),
                   lightbar.refresh()]))


def clear_diz(term):
    """ Clear file_id.diz area. """
    echo(term.move(1, browser.diz_location))
    # +2 for info line (filename, size) and empty line below it
    for i in range(browser.last_diz_len + 2):
        echo(u''.join((
            term.move(i, browser.diz_location), term.clear_eol)))


def describe_file(term, diz, directory, filename, isdir=None):
    """ Describe a file in the diz area. """
    if isdir or filename == u'..{0}'.format(os.path.sep):
        # describe directory
        description = u'{txt_Directory}: {filename}'.format(
            txt_Directory=term.bold(u'Directory'),
            filename=filename.decode('utf8'))

    else:
        # describe file
        _size = filesize(os.path.join(directory, filename))
        _filename = (filename[len(ROOT):].decode('utf8')
                     if directory == os.path.join(ROOT, FLAGGED_DIRNAME)
                     else filename.decode('utf8'))
        description = (u'{txt_Filename}: {filename}  {txt_Size}: {size}'
                       .format(txt_Filename=term.bold(u'Filename'),
                               filename=_filename,
                               txt_Size=term.bold(u'Size'),
                               size=_size))

    description = term.wrap(description, browser.max_diz_width)
    echo(u''.join((term.move(1, browser.diz_location),
                   u''.join(['{0}{1}\r\n'.format(
                             term.move_x(browser.diz_location), line)
                             for line in description]),
                   term.move(2 + len(description), browser.diz_location))))

    wrapped_diz = []
    for line in diz:
        wrapped_diz += term.wrap(line, browser.max_diz_width)
    wrapped_diz = wrapped_diz[:browser.max_diz_height - len(description) + 1]

    output = u''
    for line in wrapped_diz:
        browser.last_diz_len += 1
        output = u''.join(
            (output, term.move_x(browser.diz_location),
             line, u'\r\n'))
    echo(output)


def mark_flagged(directory, files):
    """ Add marker to flagged files. """
    files_list = list()
    for fname in files:
        prefix = u' '
        if os.path.join(directory, fname) in browser.flagged_files:
            prefix = FLAGGED_CHAR
        txt_fname = fname.strip().decode('utf8')
        item = (fname, (u'{prefix}{txt_fname}'
                        .format(prefix=prefix, txt_fname=txt_fname)))
        files_list.append(item)
    return files_list


def flagged_listdir():
    """ Build listing for flagged files pseudo-folder. """
    files = [u'{flagged_char}{txt_fname}'.format(
        flagged_char=FLAGGED_CHAR,
        txt_fname=fname[fname.rfind(os.path.sep) + 1:].decode('utf8'))
        for fname in browser.flagged_files]

    zipped_files = zip(browser.flagged_files, files)
    sorted_files = sorted(zipped_files, key=lambda x: x[1].lower())
    sorted_files.insert(0, (u'..{0}'.format(os.path.sep),
                            u' ..{0}'.format(os.path.sep)))
    return sorted_files


def regular_listdir(session, directory, sub):
    """ Build listing for regular folder. """
    files = sorted(os.listdir(directory), key=lambda x: x.lower())
    sorted_dirs = []
    sorted_files = []
    for fname in files:
        filepath = os.path.join(directory, fname)
        # skip uploads folder
        if not session.user.is_sysop and filepath == UPLOADS_DIR:
            continue
        # designate dirs with path separator suffix
        if os.path.isdir(filepath):
            sorted_dirs.append('{0}{1}'.format(fname, os.path.sep))
        else:
            sorted_files.append(fname)

    if sub:
        # we are in a subdir; add '..' parent directory entry
        sorted_dirs.insert(0, '..{0}'.format(os.path.sep))
    else:
        # we are in the root; add the flagged pseudo-folder
        sorted_dirs.insert(0, FLAGGED_DIRNAME)

    return mark_flagged(directory, sorted_dirs + sorted_files)


def reload_dir(session, directory, lightbar, sub):
    """ Reload contents of directory. """
    if directory == os.path.join(ROOT, FLAGGED_DIRNAME):
        # pseudo-folder for flagged files list
        lightbar.update(flagged_listdir())
    else:
        # actual folder
        lightbar.update(regular_listdir(session, directory, sub))


def is_flagged_dir(directory):
    """ Check to see if this is the __flagged__ directory. """
    return (directory == FLAGGED_DIRNAME or
            directory.endswith(FLAGGED_DIRNAME))


def browse_dir(session, db_desc, term, lightbar, directory, sub=False):
    """ Browse a directory. """
    # build and sort directory listing
    reload_dir(session, directory, lightbar, sub)
    echo(lightbar.refresh())
    filename, _ = lightbar.selection
    browser.last_diz_len = 0
    diz = ''
    # force it to describe the very first file when browser loads
    inp = lightbar.keyset['home'][0]
    # prime our loop
    isdir = False
    filepath = ''

    while True:
        # read from lightbar
        while not inp:
            inp = getch(0.2)
            # respond to screen dimension change by redrawing
            if session.poll_event('refresh'):
                draw_interface(term, lightbar)
                describe_file(term, diz=diz, directory=directory,
                              filename=filename, isdir=isdir)

        idx = lightbar.vitem_idx
        shift = lightbar.vitem_shift

        # pass input to lightbar
        lightbar.process_keystroke(inp)
        filename, _ = lightbar.selection

        filepath = os.path.join(directory, filename)
        relativename = filepath[len(ROOT):]
        isdir = bool(filepath[-1:] == os.path.sep)
        _, ext = os.path.splitext(filename.lower())

        if inp in lightbar.keyset['home']:
            # lightbar 'home' keystroke bug; redraw current line
            echo(lightbar.refresh_row(idx))
            echo(lightbar.refresh_row(lightbar.vitem_idx))

        elif lightbar.quit:
            # 'exit' key pressed
            return False

        elif inp in (u' ',) and filename[-1:] != os.path.sep:
            # 'flag' key pressed; don't allow flagging directories
            if filepath in browser.flagged_files:
                # already flagged; un-flag
                browser.flagged_files.remove(filepath)
            else:
                browser.flagged_files.add(filepath)

            session.user['flaggedfiles'] = browser.flagged_files
            reload_dir(session, directory, lightbar, sub)

            if is_flagged_dir(directory):
                echo(lightbar.refresh())
            else:
                echo(lightbar.refresh_row(lightbar.vitem_idx))
                lightbar.move_down()

        elif inp in (u'-',):
            # 'unflag all' pressed
            session.user['flaggedfiles'] = browser.flagged_files = set()
            reload_dir(session, directory, lightbar, sub)
            echo(lightbar.refresh())

        elif inp in (u'd',) and len(browser.flagged_files):
            download_files(term, session)
            reload_dir(session, directory, lightbar, sub)
            draw_interface(term, lightbar)

        elif inp in (u'u',):
            upload_files(term)
            reload_dir(session, directory, lightbar, sub)
            draw_interface(term, lightbar)

        elif inp in (u'e',) and session.user.is_sysop and not isdir:
            edit_description(relativename, db_desc)
            reload_dir(session, directory, lightbar, sub)
            draw_interface(term, lightbar)

        clear_diz(term)
        save_diz = True

        if lightbar.selected or inp in (term.KEY_LEFT, term.KEY_RIGHT,):

            if sub and inp is term.KEY_LEFT:
                # term.KEY_LEFT backs up
                return True

            if (isdir or is_flagged_dir(filename) and (
                    lightbar.selected or inp is term.KEY_RIGHT)):
                # 'select' key pressed

                if filename == '..{0}'.format(os.path.sep):
                    # is directory and is a parent directory; back out
                    return True

                # RECURSION
                if not browse_dir(session, db_desc, term, lightbar,
                                  filepath, True):
                    # sub directory; jump in
                    return False

                reload_dir(session, directory, lightbar, sub)
                lightbar.vitem_shift = shift
                lightbar.vitem_idx = idx
                echo(lightbar.refresh())

        if relativename in db_desc:
            save_diz = False
            diz = db_desc[relativename]
            if ext in COLLY_EXTENSIONS:
                decoder = 'cp437_art'
                if session.encoding == 'utf8':
                    decoder = COLLY_DECODING
                try:
                    diz = [line.decode(decoder, errors='replace')
                           for line in diz]
                except UnicodeEncodeError:
                    diz = [u'Invalid characters in FILE_ID.DIZ']

        elif ext in browser.diz_extractors:
            # is (supported) archive
            diz = browser.diz_extractors[ext](filepath).splitlines()

        elif ext in COLLY_EXTENSIONS:
            # is ASCII colly, pull diz from between markers if available.
            diz = get_diz_from_colly(filepath=filepath) or diz
            # save diz in raw format, but display decoded
            save_diz = False
            db_desc[relativename] = diz
            decoder = 'cp437_art'
            if session.encoding == 'utf8':
                decoder = COLLY_DECODING
            try:
                diz = [line.decode(decoder, errors='replace') for line in diz]
            except UnicodeEncodeError:
                diz = [u'Invalid characters in FILE_ID.DIZ']
                db_desc[relativename] = diz

        elif is_flagged_dir(filename):
            # is pseudo-folder for flagged files
            save_diz = False
            diz = get_instructions(term, session.user.is_sysop)

        elif isdir:
            # is directory; don't give it a description
            save_diz = False
            diz = []

        else:
            # is normal file
            save_diz = False
            diz = [u'No description']

        if not UPLOADS_DIR.find(directory) and save_diz:
            # write description to diz db when save_diz is True
            with db_desc:
                db_desc[relativename] = diz

        browser.last_diz_len = len(diz)
        describe_file(term=term, diz=diz, directory=directory,
                      filename=filename, isdir=isdir)
        echo(lightbar.refresh_quick() + lightbar.fixate())
        inp = None


def main():
    """ File browser launch point. """
    import subprocess
    import functools
    session, term = getsession(), getterminal()
    session.activity = u'Browsing files'
    db_desc = DBProxy(DIZ_DB)

    # set syncterm font, if any
    if SYNCTERM_FONT and term.kind.startswith('ansi'):
        echo(syncterm_setfont(SYNCTERM_FONT))

    # assign extractors to file types
    browser.diz_extractors['.zip'] = diz_from_zip

    # detect LHA and DMS support
    output, _ = subprocess.Popen(('which', 'lha'), stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE).communicate()
    if output:
        browser.diz_extractors['.lha'] = functools.partial(diz_from_lha,
                                                          output.rstrip())
    output, _ = subprocess.Popen(('which', 'xdms'), stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE).communicate()
    if output:
        browser.diz_extractors['.dms'] = functools.partial(diz_from_dms,
                                                          output.rstrip())

    # load flagged files
    browser.flagged_files = session.user.get('flaggedfiles', set())

    # remove missing files/dirs from flagged files, just in case
    if len(browser.flagged_files):
        for filepath in list(browser.flagged_files)[:]:
            if not os.path.exists(filepath):
                browser.flagged_files.remove(filepath)
        session.user['flaggedfiles'] = browser.flagged_files

    # fire it up!
    lightbar = Lightbar(height=term.height,
                        width=min(10, int(term.width * 0.25)),
                        xloc=0, yloc=0,
                        colors={'border': getattr(term, COLOR_BORDER),
                                'highlight': getattr(term, COLOR_HIGHLIGHT)})
    draw_interface(term, lightbar)
    with term.hidden_cursor():
        browse_dir(session, db_desc, term, lightbar, ROOT)
    echo(term.move(term.height, term.width))
    echo(u'\r\n\r\n' + term.normal)



"""
SFTP server for x/84.

In order to configure x/84 as an SFTP server, you will need to add an [sftp]
section to your default.ini file and add a `root` option. This points to the
location of your SFTP server's root directory. Within that directory, ensure
that there is a directory named `__uploads__`.

This is based on paramiko's `StubSFTPServer` implementation.
"""

# std imports
import logging
import os

# 3rd-party
from paramiko import (
    SFTPServerInterface,
    SFTPServer,
    SFTPAttributes,
    SFTPHandle,
    SFTP_OK,
    SFTP_PERMISSION_DENIED,
)

# directory name for flagged files
flagged_dirname = '__flagged__'
uploads_dirname = '__uploads__'


class X84SFTPHandle(SFTPHandle):

    """ SFTP File handler for x/84. """

    def __init__(self, *args, **kwargs):
        """ Class initializer. """
        self.log = logging.getLogger(__name__)
        self.user = kwargs.pop('user')
        SFTPHandle.__init__(self, *args, **kwargs)

    def stat(self):
        """ Stat the file descriptor. """
        self.log.debug('stat')
        try:
            return SFTPAttributes.from_stat(os.fstat(self.readfile.fileno()))
        except OSError as err:
            return SFTPServer.convert_errno(err.errno)

    def chattr(self, attr):
        """ Change attributes of the file descriptor. """
        if not self.user.is_sysop:
            return SFTP_PERMISSION_DENIED
        self.log.debug('chattr ({0!r})'.format(attr))
        # python doesn't have equivalents to fchown or fchmod, so we have to
        # use the stored filename
        try:
            SFTPServer.set_file_attr(self.filename, attr)
            return SFTP_OK
        except OSError as err:
            return SFTPServer.convert_errno(err.errno)


class X84SFTPServer(SFTPServerInterface):

    """ SFTP server for x/84. """

    def __init__(self, *args, **kwargs):
        """ Class initializer. """
        from x84.bbs import get_ini
        self.log = logging.getLogger(__name__)

        # root file folder,
        self.root = get_ini(section='sftp', key='root')

        # default file mode for uploaded files,
        _base = 8  # (value is octal!)
        self.mode = int(get_ini(
            section='sftp', key='uploads_filemode') or '644', _base)

        # allow anonymous login where enabled, otherwise use the
        # given `username' authenticated by ssh
        from x84.bbs.userbase import get_user, User
        _ssh_session = kwargs.pop('ssh_session')
        self.user = (User(u'anonymous') if _ssh_session.anonymous
                     else get_user(_ssh_session.username))
        self.flagged = set()

        SFTPServerInterface.__init__(self, *args, **kwargs)

    def _dummy_dir_stat(self):
        """ Stat on our dummy __flagged__ directory. """
        self.log.debug('_dummy_dir_stat')
        attr = SFTPAttributes.from_stat(
            os.stat(self.root))
        attr.filename = flagged_dirname
        return attr

    def _realpath(self, path):
        """ Get the real path of a given path. """
        self.log.debug('_realpath({0!r})'.format(path))
        if path.endswith(flagged_dirname):
            self.log.debug('fake dir path: {0!r}'.format(path))
            return self.root + path
        elif path.find(flagged_dirname) > -1:
            self.log.debug('fake file path: {0!r}'.format(path))
            for fname in self.flagged:
                fstripped = fname[fname.rindex(os.path.sep) + 1:]
                pstripped = path[path.rindex('/') + 1:]
                if fstripped == pstripped:
                    self.log.debug('file is actually {0}'.format(fname))
                    return fname

        # pylint: disable=E1101
        #         Instance of 'X84SFTPServer' has no 'canonicalize' member
        return self.root + self.canonicalize(path)

    def _is_uploaddir(self, path):
        """ Check if this is the upload directory. """
        return (path == '/{0}'.format(uploads_dirname))

    def list_folder(self, path):
        """ List contents of a folder. """
        self.log.debug('list_folder({0!r})'.format(path))
        rpath = self._realpath(path)
        if not self.user.is_sysop and self._is_uploaddir(path):
            return []
        try:
            out = []
            if path == u'/':
                out.append(self._dummy_dir_stat())
            elif flagged_dirname in path:
                self.flagged = self.user.get('flaggedfiles', set())
                for fname in self.flagged:
                    rname = fname
                    attr = SFTPAttributes.from_stat(os.stat(rname))
                    attr.filename = fname[fname.rindex('/') + 1:]
                    out.append(attr)
                return out
            flist = os.listdir(rpath)
            for fname in flist:
                attr = SFTPAttributes.from_stat(
                    os.stat(os.path.join(rpath, fname)))
                attr.filename = fname
                out.append(attr)
            return out
        except OSError as err:
            return SFTPServer.convert_errno(err.errno)

    def stat(self, path):
        """ Stat a given path. """
        self.log.debug('stat({0!r})'.format(path))
        if path.endswith(flagged_dirname):
            return self._dummy_dir_stat()
        elif path.find(flagged_dirname) > -1:
            for fname in self.flagged:
                fstripped = fname[fname.rindex(os.path.sep) + 1:]
                pstripped = path[path.rindex('/') + 1:]
                if fstripped == pstripped:
                    self.log.debug('file is actually {0}'.format(fname))
                    return SFTPAttributes.from_stat(fname)
        path = self._realpath(path)
        try:
            return SFTPAttributes.from_stat(os.stat(path))
        except OSError as err:
            return SFTPServer.convert_errno(err.errno)

    def lstat(self, path):
        """ Lstat a given path. """
        self.log.debug('lstat({0!r})'.format(path))
        if path.endswith(flagged_dirname):
            return self._dummy_dir_stat()
        elif path.find(flagged_dirname) > -1:
            for fname in self.flagged:
                fstripped = fname[fname.rindex(os.path.sep) + 1:]
                pstripped = path[path.rindex('/') + 1:]
                if fstripped == pstripped:
                    self.log.debug('file is actually {0}'.format(fname))
                    return SFTPAttributes.from_stat(fname)
        path = self._realpath(path)
        try:
            return SFTPAttributes.from_stat(os.lstat(path))
        except OSError as err:
            return SFTPServer.convert_errno(err.errno)

    def open(self, path, flags, attr):
        """ Up/download the given path. """
        self.log.debug('lstat({0!r}, {1!r}, {2!r})'
                       .format(path, flags, attr))
        path = self._realpath(path)
        if (flags & os.O_CREAT and (uploads_dirname not in path and
                                    not self.user.is_sysop) or
                (uploads_dirname in path and os.path.exists(path))):
            return SFTP_PERMISSION_DENIED
        try:
            binary_flag = getattr(os, 'O_BINARY', 0)
            flags |= binary_flag
            filedesc = os.open(path, flags, self.mode)
        except OSError as err:
            return SFTPServer.convert_errno(err.errno)
        if (flags & os.O_CREAT) and (attr is not None):
            attr._flags &= ~attr.FLAG_PERMISSIONS
            SFTPServer.set_file_attr(path, attr)
        if flags & os.O_WRONLY:
            if flags & os.O_APPEND:
                fstr = 'ab'
            else:
                fstr = 'wb'
        elif flags & os.O_RDWR:
            if flags & os.O_APPEND:
                fstr = 'a+b'
            else:
                fstr = 'r+b'
        else:
            # O_RDONLY (== 0)
            fstr = 'rb'
        try:
            openfile = os.fdopen(filedesc, fstr)
        except OSError as err:
            return SFTPServer.convert_errno(err.errno)
        fobj = X84SFTPHandle(flags, user=self.user)
        fobj.filename = path
        fobj.readfile = openfile
        fobj.writefile = openfile

        if path in self.flagged:
            self.flagged.remove(path)
            self.user['flaggedfiles'] = self.flagged
        return fobj

    def remove(self, path):
        """ Delete the given path. """
        if not self.user.is_sysop or flagged_dirname in path:
            return SFTP_PERMISSION_DENIED
        self.log.debug('remove({0!r})'.format(path))
        path = self._realpath(path)
        try:
            os.remove(path)
        except OSError as err:
            return SFTPServer.convert_errno(err.errno)
        return SFTP_OK

    def rename(self, oldpath, newpath):
        """ Rename the ``oldpath`` to ``newpath``. """
        if not self.user.is_sysop or flagged_dirname in oldpath or \
                flagged_dirname in newpath:
            return SFTP_PERMISSION_DENIED
        self.log.debug('rename({0!r}, {1!r})'.format(oldpath, newpath))
        oldpath = self._realpath(oldpath)
        newpath = self._realpath(newpath)
        try:
            os.rename(oldpath, newpath)
        except OSError as err:
            return SFTPServer.convert_errno(err.errno)
        return SFTP_OK

    def mkdir(self, path, attr):
        """ Makea  new directory. """
        if not self.user.is_sysop or flagged_dirname in path:
            return SFTP_PERMISSION_DENIED
        self.log.debug('mkdir({0!r}, {1!r})'.format(path, attr))
        path = self._realpath(path)
        try:
            os.mkdir(path)
            if attr is not None:
                SFTPServer.set_file_attr(path, attr)
        except OSError as err:
            return SFTPServer.convert_errno(err.errno)
        return SFTP_OK

    def rmdir(self, path):
        """ Remove a directory. """
        if not self.user.is_sysop or flagged_dirname in path:
            return SFTP_PERMISSION_DENIED
        self.log.debug('rmdir({0!r})'.format(path))
        path = self._realpath(path)
        try:
            os.rmdir(path)
        except OSError as err:
            return SFTPServer.convert_errno(err.errno)
        return SFTP_OK

    def chattr(self, path, attr):
        """ Change path's attributes. """
        if self._is_uploaddir(path):
            return SFTP_PERMISSION_DENIED
        elif not self.user.is_sysop or \
                uploads_dirname in path or \
                flagged_dirname in path:
            return SFTP_PERMISSION_DENIED
        self.log.debug('chattr({0!r})'.format(path))
        path = self._realpath(path)
        try:
            SFTPServer.set_file_attr(path, attr)
        except OSError as err:
            return SFTPServer.convert_errno(err.errno)
        return SFTP_OK

    def symlink(self, target_path, path):
        """ Create symbolic link. """
        if not self.user.is_sysop or flagged_dirname in path:
            return SFTP_PERMISSION_DENIED
        self.log.debug('symlink({0!r}, {1!r})'.format(target_path, path))
        path = self._realpath(path)
        if (len(target_path) > 0) and (target_path[0] == '/'):
            # absolute symlink
            target_path = os.path.join(self.root, target_path[1:])
            if target_path[:2] == '//':
                # bug in os.path.join
                target_path = target_path[1:]
        else:
            # compute relative to path
            abspath = os.path.join(os.path.dirname(path), target_path)
            if abspath[:len(self.root)] != self.root:
                # this symlink isn't going to work anyway
                # -- just break it immediately
                target_path = '<error>'
        try:
            os.symlink(target_path, path)
        except OSError as err:
            return SFTPServer.convert_errno(err.errno)
        return SFTP_OK

    def readlink(self, path):
        """ Read symbolic link. """
        self.log.debug('readlink({0!r})'.format(path))
        path = self._realpath(path)
        try:
            symlink = os.readlink(path)
        except OSError as err:
            return SFTPServer.convert_errno(err.errno)
        # if it's absolute, remove the root
        if os.path.isabs(symlink):
            if symlink[:len(self.root)] == self.root:
                symlink = symlink[len(self.root):]
                if (len(symlink) == 0) or (symlink[0] != '/'):
                    symlink = '/' + symlink
            else:
                symlink = '<error>'
        return symlink



""" Session IPC package for x/84. """
# std imports
import logging

# local
from x84.bbs.session import getsession


def make_root_logger(out_queue):
    """
    Remove and re-address the root logging handler.

    Any existing handlers of the current process are removed and
    the root logger is re-address to send via an IPC output event
    queue.
    """
    root = logging.getLogger()
    map(root.removeHandler, root.handlers)
    root.addHandler(IPCLogHandler(out_queue=out_queue))


class IPCLogHandler(logging.Handler):

    """
    Log handler that sends the log up the 'event pipe'.

    This is a rather novel solution that seems overlooked in documentation,
    a forked process must have some method to propagate its logging records
    up through the main process, otherwise they are lost.
    """

    def __init__(self, out_queue):
        """ Constructor method, requires multiprocessing.Pipe. """
        logging.Handler.__init__(self)
        self.oqueue = out_queue

    def emit(self, record):
        """ Emit log record via IPC output queue. """
        try:
            e_inf = record.exc_info
            if e_inf:
                # a strange side-effect,
                # sets record.exc_text
                dummy = self.format(record)  # NOQA
                record.exc_info = None
            record.handle = None
            session = getsession()
            if session:
                record.handle = session.user.handle
            self.oqueue.send(('logger', record))
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            self.handleError(record)


class IPCStream(object):

    """
    Connect blessed.Terminal argument 'stream' to 'writer' queue.

    The ``writer`` queue is a ``multiprocessing.Pipe`` whose master-side
    is polled for output in x84.engine.  Only the ``write()`` method of
    this "stream" and ``is_a_tty`` attribute is called or evaluated by
    blessed.Terminal.  The attribute ``is_a_tty`` is mocked as ``True``.
    """

    def __init__(self, writer):
        self.writer = writer
        self.is_a_tty = True

    def write(self, ucs, encoding='ascii'):
        """
        Sends unicode text to Pipe.

        Default encoding is 'ascii', which is unset only when used
        with blessings, which rarely writes directly to the stream
        (context managers, such as "with term.location(0, 0):" have
        such side effects).
        """
        # wrap 'ucs' with call to 'unicode()', so that special unicode
        # instances such as blessed.formatters.ParameterizingProxyString
        # can be pickled -- as this one in particular contains a local
        # function (lambda) as an attribute -- which would fail:
        # PicklingError: Can't pickle <type 'function'>: attribute
        #                lookup __builtin__.function failed
        self.writer.send(('output', (unicode(ucs), encoding)))



""" telnet client for x/84 """
KEY_POLL = 0.015
IS = chr(0)
SEND = chr(1)


def main(host, port=None, encoding='cp437'):
    """
    Call script with argument host and optional argument port to connect to a
    telnet server. ctrl-^ to disconnect.
    """
    # pylint: disable=R0914,R0912,R0915
    #         Too many local variables
    #         Too many branches
    #         Too many statements
    import telnetlib
    from functools import partial
    from x84.bbs import getsession, getterminal, echo, getch, from_cp437, telnet
    import logging
    log = logging.getLogger()

    assert encoding in ('utf8', 'cp437')
    session, term = getsession(), getterminal()
    session.activity = 'connecting to %s' % (host,)
    port = int(port) if port is not None else 23
    telnet_client = telnetlib.Telnet()
    telnet_client.set_option_negotiation_callback(partial(
        telnet.callback_cmdopt, env_term=session.env['TERM'], height=term.height, width=term.width))
    echo(u"\r\n\r\nEscape character is 'ctrl-^.'")
    if not session.user.get('expert', False):
        getch(3)
    echo(u'\r\nTrying %s:%s... ' % (host, port,))
    # pylint: disable=W0703
    #         Catching too general exception Exception
    try:
        telnet_client.open(host, port)
    except Exception as err:
        echo(term.bold_red('\r\n%s\r\n' % (err,)))
        echo(u'\r\n press any key ..')
        getch()
        return

    echo(u'\r\n... ')
    inp = session.read_event('input', timeout=0)
    echo(u'\r\nConnected to %s.' % (host,))
    session.activity = 'connected to %s' % (host,)
    carriage_returned = False
    with term.fullscreen():
        while True:
            if encoding == 'cp437':
                try:
                    unistring = from_cp437(
                        telnet_client.read_very_eager().decode('iso8859-1'))
                except EOFError:
                    break
            else:
                unistring = telnet_client.read_very_eager().decode('utf8')
            if 0 != len(unistring):
                echo(unistring)
            if inp is not None:
                if inp == chr(30):  # ctrl-^
                    telnet_client.close()
                    echo(u'\r\n' + term.clear_el + term.normal)
                    break
                elif not carriage_returned and inp in (b'\x0d', b'\x0a'):
                    telnet_client.write(b'\x0d')
                    log.debug('send {!r}'.format(b'\x0d'))
                    carriage_returned = True
                elif carriage_returned and inp in (b'\x0a', b'\x00'):
                    carriage_returned = False
                elif inp is not None:
                    telnet_client.write(inp)
                    log.debug('send {!r}'.format(inp))
                    carriage_returned = False
            inp = session.read_event('input', timeout=KEY_POLL)
    echo(u'\r\nConnection closed.\r\n')
    echo(u''.join(('\r\n\r\n', term.clear_el, term.normal, 'press any key')))
    echo(u'\x1b[r')  # unset 'set scrolling region', sometimes set by BBS's
    session.flush_event('input')
    getch()
    return
